#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
common.py
Python 3.9+
Date created: June 14th, 2020
Date modified: July 3rd, 2023
"""

import platform
import sqlite3
import sys
from enum import Enum

VERSION = "0.6.1"


class Platform(Enum):
    """An enum used to indicate the running operating system
    """

    OTHER = 0
    LINUX = 1
    MAC = 2
    WINDOWS = 3


def get_platform() -> str:
    """Get the current operating system.

    :return: the operating system this program is running on
    """
    system = platform.system()
    if system == "Linux":
        return str(Platform.LINUX.name)
    if system == "Darwin":
        return str(Platform.MAC.name)
    if system == "Windows":
        return str(Platform.WINDOWS.name)
    raise NotImplementedError(f"Platform {system} is not supported yet!")


def fetch_db_data(db, command) -> list:
    """Send queries to the sqlite database and return the results.

    :param db: the path to the database
    :param command: the SQL command
    :return: data from the database
    """
    try:
        conn = sqlite3.connect(db)
        cur = conn.cursor()
        cur.execute(command)
        return cur.fetchall()
    except Exception as e:
        sys.exit(f"Error reading the database: {e}\nDid you specify the correct path?")



#!/usr/bin/env python3

"""
Data Container
Python 3.9+
Author: niftycode
Modified by: -
Date created: February 19th, 2022
Date modified: -
"""

from dataclasses import dataclass


@dataclass
class MessageData:
    """This dataclass is the store for the data:
    user id, text, date, service and account (caller id).
    """

    user_id: str
    text: str
    date: str
    date_read: str
    service: str
    account: str
    is_from_me: int

    def __str__(self):
        """String representation

        :return: the representation of this object
        """
        return (
            f"user id:\t\t{self.user_id}\n"
            f"date and time:\t\t{self.date}\n"
            f"service:\t\t{self.service}\n"
            f"caller id:\t\t{self.account}\n"
            f"is_from_me:\t\t{self.is_from_me}\n"
            f"\n"
            f"text:\n"
            f"=====\n"
            f"{self.text}\n"
            f"\n"
            f"----------------------------------------------------------------\n"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
cli.py
Entrypoint to the command line interface.
Python 3.9+
Date created: October 15th, 2020
Date modified: June 2nd, 2023
"""

import argparse
import os
import sys

from os.path import expanduser

from imessage_reader import fetch_data
from imessage_reader import info


# Path to the chat.db file on macOS
# Note: This path is used if the user does not specify a path.
MACOS_DB_PATH = expanduser("~") + "/Library/Messages/chat.db"


def get_parser() -> argparse.ArgumentParser:
    """Create the argument parser

    :rtype: object
    """
    parser = argparse.ArgumentParser(
        description="A tool to fetch messages from the chat.db file."
    )

    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs="?",
        const=MACOS_DB_PATH,
        default=MACOS_DB_PATH,
        help="Path to the database file"
    )

    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        default="nothing",
        help="Specify the output: e => Excel, s => SQLite"
    )

    parser.add_argument(
        "-r",
        "--recipients",
        help="Show the recipients",
        action="store_true"
    )

    parser.add_argument(
        "-v",
        "--version",
        help="Show the current version.",
        action="store_true"
    )

    return parser


def check_database_path(args):
    """ Parse arguments from sys.argv and invoke the evaluate method.

    :param args: the user's input
    """
    if args.path == MACOS_DB_PATH:
        evaluate(MACOS_DB_PATH, args.output, args.recipients, args.version)
    elif os.path.isdir(args.path):
        db_path = args.path + "/chat.db"
        evaluate(db_path, args.output, args.recipients, args.version)
    else:
        sys.exit("Path doesn't exist! Exit program.")


def evaluate(path: str, output: str, recipients: bool, version: bool):
    """Evaluate the given options and perform the appropriate actions.

    :param path: path to the chat.db file
    :param output: create an Excel/SQLite3 file
    :param recipients: recipients of the messages
    :param version: specify if the version of this program should be shown
    """
    data = fetch_data.FetchData(path)

    if version:
        info.app_info()
        sys.exit()

    if recipients:
        data.show_user_txt("recipients")
        sys.exit()

    if output == "e" or output == "excel":
        data.show_user_txt("excel")
    elif output == "s" or output == "sqlite" or output == "sqlite3":
        data.show_user_txt("sqlite")
    else:
        data.show_user_txt("nothing")


def main():
    """Entrypoint to the command-line interface (CLI).
    """
    parser = get_parser()
    args = parser.parse_args()
    check_database_path(args)


if __name__ == "__main__":
    main()



# db.py 

# Connect to
import sqlite3
import os 
import datetime
import math
import pytz

# Works for mac Catalina
home = os.environ['HOME']
db_path = f'{home}/Library/Messages/chat.db'

db = None

def open():
    global db
    if db:
        return db
    # Read only mode
    db = sqlite3.connect(db_path, uri=True)

def clean_up():
    if db:
        db.close()

DATE_OFFSET = 978307200

def apple_time_now():
     return math.floor(datetime.datetime.now() / 1000) - DATE_OFFSET

def from_apple_time(ts):
    if ts==0:
        return None

    if unpack_time(ts) != 0:
        ts = unpack_time(ts)
    
    return datetime.datetime.fromtimestamp((ts + DATE_OFFSET), tz=pytz.timezone('US/Pacific'))

def unpack_time(ts):
    return math.floor(ts / (10**9))

def pack_time_conditionally(ts):
    return ts * 10**9

def get_most_recently_sent_text():
    return db.execute("""SELECT guid, id as handle, text, date, date_read, date_delivered
    FROM message
    LEFT OUTER JOIN handle ON message.handle_id=handle.ROWID
    WHERE is_from_me = 1 
    ORDER BY date DESC
    LIMIT 1""").fetchone()[0]

def get_message(guid): 
    message = db.execute(f"""SELECT guid, date, date_read, date_delivered
    FROM message
    LEFT OUTER JOIN handle ON message.handle_id=handle.ROWID
    WHERE is_from_me = 1 and guid="{guid}"
    LIMIT 1""").fetchone()

    return {
        'guid': message[0],
        'date': from_apple_time(message[1]),
        'date_read': from_apple_time(message[2]),
        'date_delivered': from_apple_time(message[3])
        }


#!/usr/bin/env python3

"""
Write Excel file containing iMessage data (user id, text, date, service, account)
Python 3.9+
Date created: October 1st, 2020
Date modified: August 28th, 2021
"""

from datetime import datetime

import openpyxl
from openpyxl.styles import Font


class ExelWriter:
    """This class manages the export to excel.
    """

    def __init__(self, imessage_data: list, file_path: str):
        """Constructor method

        :param imessage_data: list with MessageData objects
                containing user id, text, date, service and account
        :param file_path: path to the location of the Excel file
        """
        self.imessage_data = imessage_data
        self.file_path = file_path

    def write_data(self):
        """Write data to Excel
        (user id, text, date, service, account, is_from_me)
        """

        users = []
        messages = []
        dates = []
        services = []
        accounts = []
        is_from_me = []

        for data in self.imessage_data:
            users.append(data.user_id)
            messages.append(data.text)
            dates.append(data.date)
            services.append(data.service)
            accounts.append(data.account)
            is_from_me.append(data.is_from_me)

        # Call openpyxl.Workbook() to create a new blank Excel workbook
        workbook = openpyxl.Workbook()

        # Activate a sheet
        sheet = workbook.active

        # Set a title
        sheet.title = "iMessages"

        # Set headline style
        bold16font = Font(size=16, bold=True)

        sheet["A1"] = "User ID"
        sheet["A1"].font = bold16font

        sheet["B1"] = "Message"
        sheet["B1"].font = bold16font

        sheet["C1"] = "Date"
        sheet["C1"].font = bold16font

        sheet["D1"] = "Service"
        sheet["D1"].font = bold16font

        sheet["E1"] = "Destination Caller ID"
        sheet["E1"].font = bold16font

        sheet["F1"] = "Is From Me"
        sheet["F1"].font = bold16font

        # Write users to 1st column
        users_row = 2
        for user in users:
            sheet.cell(row=users_row, column=1).value = user
            users_row += 1

        # Write messages to 2nd column
        messages_row = 2
        for message in messages:
            sheet.cell(row=messages_row, column=2).value = message
            messages_row += 1

        # Write date to 3rd column
        dates_row = 2
        for date in dates:
            sheet.cell(row=dates_row, column=3).value = date
            dates_row += 1

        # Write services to 4th column
        service_row = 2
        for service in services:
            sheet.cell(row=service_row, column=4).value = service
            service_row += 1

        # Write accounts to 5th column
        account_row = 2
        for account in accounts:
            sheet.cell(row=account_row, column=5).value = account
            account_row += 1

        # Write is_from_me to 6th column
        is_from_me_row = 2
        for from_me in is_from_me:
            sheet.cell(row=is_from_me_row, column=6).value = from_me
            is_from_me_row += 1

        # Save the workbook (Excel file)
        try:
            workbook.save(
                self.file_path + f'iMessage-Data_{datetime.now().strftime("%Y-%m-%d")}.xlsx'
            )
            print()
            print(">>> Excel file successfully created! <<<")
            print("You find the Excel file in your Documents folder.")
            print()
        except IOError as e:
            print(">>> Cannot write Excel file! <<<")
            print(e)



from datetime import datetime

from .imessage_reader import fetch_data
from .py_imessage import imessage

from .mac_db import get_db_path


class MessageClient:
    IMESSAGE_CREATION = '2011-01-01 00:00:00'

    def __init__(self, *args, **kwargs):
        # A time filter to ensure messages being read come after a certain date
        self.time_filter = kwargs.get("time_filter") or self.IMESSAGE_CREATION
        if type(self.time_filter) is str:
            self.time_filter = self._parse_time(self.time_filter)

        # The location of the chat.db file used by iMessage
        self.db_path = kwargs.get("db_path", get_db_path())

        # The function invoked to handle new messages
        if "handle_message" in kwargs:
            self.handle_message = kwargs["handle_message"]

        if "handle_post_read" in kwargs:
            self.post_read_and_handle = kwargs["handle_post_read"]

        self.read_messages_cache = dict()

    def _hash_message_for_cache(self, message):
        return hash(f"{message[0]} :: {message[1]} :: {message[2]}")

    def _parse_time(self, dt):
        return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")

    def _get_messages(self, no_filter=False):
        fd = fetch_data.FetchData(self.db_path)
        my_data = fd.get_messages()

        if no_filter:
            return my_data

        return [m for m in my_data if self._parse_time(m[2]) > self.time_filter]

    def _get_inbound_messages(self, no_filter=False):
        messages = self._get_messages(no_filter)
        return [m for m in messages if m[-1] != 1]

    def _number_requested_stop(self, number):
        messages = self._get_inbound_messages(no_filter=True)
        return (number in m[0] and m[1] == "STOP" for m in messages)

    def send_message(self, to, content, ignore_stop=False):
        if not ignore_stop and self._number_requested_stop(to):
            # If you want to ship this to the iOS app store then you'll need to stop when the user requests so
            return

        imessage.send(to, content)

    def handle_message(self, from_number, body):
        raise NotImplemented

    def post_read_and_handle(self):
        # Optional, can be overwritten by kwargs["handle_post_read"]
        pass

    def read_and_handle(self):
        messages = self._get_inbound_messages()
        for m in messages:
            h = self._hash_message_for_cache(m)
            if h not in self.read_messages_cache:
                self.read_messages_cache[h] = True
                from_number = m[0]
                body = m[1]
                possible_reply = self.handle_message(from_number, body)
                if possible_reply is not None:
                    self.send_message(from_number, possible_reply)

        self.post_read_and_handle()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Create a SQLite3 database containing iMessage data (user id, text, date, service)
Python 3.9+
Date created: April 30th, 2021
Date modified: August 28th, 2021
"""

import sqlite3


class CreateDatabase:
    """This class manages the export to SQLite.
    """

    def __init__(self, imessage_data: list, file_path: str):
        """Constructor method

        :param imessage_data: list with MessageData objects
                containing user id, text, date, service and account
        :param file_path: the path to the location of the Excel file
        """
        self.imessage_data = imessage_data
        self.file_path = file_path

    def create_sqlite_db(self):
        """Create a SQLite3 database in the Desktop folder.
        Add user, text, date and service to the database.
        """
        database = self.file_path + "iMessage-Data.sqlite"

        conn = sqlite3.connect(database)
        cur = conn.cursor()

        cur.execute("DROP TABLE IF EXISTS Messages")

        cur.execute(
            """
        CREATE TABLE IF NOT EXISTS Messages (user_id TEXT,
        message TEXT,
        date TEXT,
        service TEXT,
        destination_caller_id TEXT, 
        is_from_me TEXT)"""
        )

        for data in self.imessage_data:
            cur.execute(
                """INSERT INTO Messages (user_id, message, date, service, destination_caller_id, is_from_me)
                VALUES(?, ?, ?, ?, ?, ?)""",
                (
                    data.user_id,
                    data.text,
                    data.date,
                    data.service,
                    data.account,
                    data.is_from_me,
                ),
            )

        conn.commit()
        cur.close()

        print()
        print(">>> SQLite database successfully created! <<<")
        print("You find the Database in your Documents folder.")
        print()



#!/usr/bin/env python3

"""
App info, license, version
Python 3.9+
Date created: June 14th, 2020
Date modified: February 22nd, 2022
"""


from imessage_reader import common

VERSION = common.VERSION


def app_info():
    """Show app infos (Version & License)
    """
    print("\n\n    ##### A Python script to read iMessage data #####")
    print("    #              Created by niftycode             #")
    print(f"    #                 Version {VERSION}                 #")
    print("    #                  MIT License                  #")
    print("    #################################################\n\n")
    print()



from . import db_conn
import os
import subprocess
import platform
from time import sleep
from shlex import quote

import sys
if sys.version_info[0] < 3:
    print("Must be using Python 3")

print(platform.mac_ver())

def _check_mac_ver():
    mac_ver, _, _ = platform.mac_ver()
    mac_ver = float('.'.join(mac_ver.split('.')[:2]))
    if mac_ver >= 10.16:
        print(mac_ver)
    else:
        print("outdated OS")
    return mac_ver
        

def send(phone, message):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    relative_path = 'osascript/send_message.js'
    path = f'{dir_path}/{relative_path}'
    cmd = f'osascript -l JavaScript {path} {quote(phone)} {quote(message)}'
    subprocess.call(cmd, shell=True)

    # Get chat message db that was sent (look at most recently sent to a phone number)
    db_conn.open()
    
    # Option 1: Loop until result is valid (hard to determine validity without adding other info to the DB)
    # Option 2: Sleep for 1 sec to allow local db to update :((
    sleep(1)
    guid = db_conn.get_most_recently_sent_text()
    return guid


def status(guid):
    db_conn.open()
    message = db_conn.get_message(guid)
    return message


def check_compatibility(phone):
    mac_ver = _check_mac_ver()
    is_imessage = False
    
    dir_path = os.path.dirname(os.path.realpath(__file__))
    relative_path = 'osascript/check_imessage.js'
    path = f'{dir_path}/{relative_path}'
    cmd = f'osascript -l JavaScript {path} {phone} {mac_ver}'
    # Gets all the output from the imessage
    output = subprocess.check_output(cmd, shell=True)

    if 'true' in output.decode('utf-8'):
        is_imessage = True

    return is_imessage



from .message_client import MessageClient


def read_messages(since_time=None, inbound_only=True):
    mc = MessageClient(
        time_filter=since_time
    )

    if inbound_only:
        return mc._get_inbound_messages()

    return mc._get_messages()



from datetime import datetime
from requests import post
from time import sleep

from py_imessage import imessage
import mac_imessage
from imessage_reader import fetch_data

from .mac_db import get_db_path


DB_PATH = get_db_path()
UNREAD_STRING = '2000-12-31 19:00:00'
TODAY = '2024-08-13 22:00:00'

def parse_time(dt):
  return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")

def read_messages(time_filter):
  global DB_PATH
  global UNREAD_STRING
  global TODAY

  fd = fetch_data.FetchData(DB_PATH)
  my_data = fd.get_messages()

  current_time = datetime.now()
  return [m for m in my_data if parse_time(m[2]) > time_filter]

def has_messaged_previously_target(to, target):
  all_messages = read_messages(parse_time(TODAY))
  return any(to in message[0] and message[1] == target for message in all_messages)

def has_messaged_previously(to):
  all_messages = read_messages(parse_time(TODAY))
  return any(to in message[0] for message in all_messages)

def send_message(to, content):
  if to != "4158667579" and has_messaged_previously_target(to, "STOP"):
    print("We are not going to send a message because said to stop")
    return # Someone doesn't want to receive messages
  imessage.send(to, content)

def foo():
    print("Hello world")

if __name__ == "__main__":
    foo()



from datetime import datetime
from time import sleep

from .message_client import MessageClient


def create_simple_message_client(message_handler, interval=1):
    time_of_run = datetime.now()
    
    mc = MessageClient(
        handle_message=message_handler,
        time_filter=time_of_run,
    )

    while True:
        sleep(interval)

        try:
            mc.read_and_handle()
        except Exception as e: # Very failure tolerant
            print("Exception!")
            print(e)



#!/usr/bin/env python3

"""
Fetch data, print data, show recipients or export data.
Python 3.9+
Author: niftycode
Modified by: thecircleisround
Date created: October 8th, 2020
Date modified: June 27th, 2023
"""

import sys
import logging

from os.path import expanduser

from . import common, create_sqlite, write_excel, data_container


# logging.basicConfig(level=logging.DEBUG)
logging.basicConfig(level=logging.INFO)


# noinspection PyMethodMayBeStatic
class FetchData:
    """
    This class contains the methods to fetch,
    print and export the messages.
    """

    # The SQL command
    SQL_CMD = (
        "SELECT "
        "text, "
        "datetime((date / 1000000000) + 978307200, 'unixepoch', 'localtime'),"
        "handle.id, "
        "handle.service, "
        "message.destination_caller_id, "
        "message.is_from_me, "
        "message.attributedBody, "
        "message.cache_has_attachments, "
        "datetime((date_read / 1000000000) + 978307200, 'unixepoch', 'localtime')"
        "FROM message "
        "JOIN handle on message.handle_id=handle.ROWID"
    )

    def __init__(self, db_path: str, system=None):
        """Constructor method

        :param db_path: Path to the chat.db file
        :param system: Operating System
        """
        self.db_path = db_path
        if system is None:
            self.operating_system = common.get_platform()

    def _check_system(self):
        # TODO: Change this later
        if self.operating_system == "WINDOWS":
            sys.exit("Your operating system is not supported yet!")

    def _read_database(self) -> list:
        """Fetch data from the database and store the data in a list.

        :return: list containing the user id, messages, the service and the account
        """

        rval = common.fetch_db_data(self.db_path, self.SQL_CMD)

        data = []
        for row in rval:
            text = row[0]
            if row[7] == 1:
                pass
                # text = "<Message with no text, but an attachment.>"
            # the chat.db has some weird behavior where sometimes the text value is None
            # and the text string is buried in a binary blob under the attributedBody field.
            if text is None and row[6] is not None:
                try:
                    text = row[6].split(b"NSString")[1]
                    text = text[
                        5:
                    ]  # stripping some preamble which generally looks like this: b'\x01\x94\x84\x01+'

                    # this 129 is b'\x81, python indexes byte strings as ints,
                    # this is equivalent to text[0:1] == b'\x81'
                    if text[0] == 129:
                        length = int.from_bytes(text[1:3], "little")
                        text = text[3: length + 3]
                    else:
                        length = text[0]
                        text = text[1: length + 1]
                    text = text.decode()

                    logging.debug(text)

                except Exception as e:
                    print(e)
                    sys.exit("ERROR: Can't read a message.")

            # print("We are working with the following row")
            # print(row)
            # print(f"There are {len(row)} items in the above row")
            data.append(
                data_container.MessageData(row[2], text, row[1], row[8], row[3], row[4], row[5])
            )

        return data

    def show_user_txt(self, export: str):
        """Invoke _read_database(), print fetched data and export data.
        (This method is for CLI usage.)

        :param export: Determine whether to export data
        """

        # Check the running operating system
        self._check_system()

        # Read chat.db
        fetched_data = self._read_database()

        # CLI output
        if export == "nothing":
            for data in fetched_data:
                print(data)

        # Excel export
        if export == "excel":
            self._export_excel(fetched_data)

        # SQLite3 export
        if export == "sqlite":
            self._export_sqlite(fetched_data)

        # Show all recipients
        if export == "recipients":
            self._get_recipients()

    def _export_excel(self, data: list):
        """Export data (write Excel file)

        :param data: message objects containing user id, message, date, service, account
        """
        file_path = expanduser("~") + "/Documents/"
        # TODO: Exception handling
        ew = write_excel.ExelWriter(data, file_path)
        ew.write_data()

    def _export_sqlite(self, data: list):
        """Export data (create SQLite3 database).

        :param data: message objects containing user id, message, date, service, account
        """
        file_path = expanduser("~") + "/Documents/"
        # TODO: Exception handling
        cd = create_sqlite.CreateDatabase(data, file_path)
        cd.create_sqlite_db()

    def _get_recipients(self):
        """Create a list containing all recipients and
        show the recipients in the command line.
        """
        fetched_data = self._read_database()

        # Create a list with recipients
        recipients = [i.user_id for i in fetched_data if i.is_from_me == 0]

        print()
        print("List of Recipients")
        print("------------------------")
        print()

        for recipient in recipients:
            print(recipient)

    def get_messages(self) -> list:
        """Create a list with tuples (user id, message, date, service, account, is_from_me)
        (This method is for module usage.)

        :return: list with tuples (user id, message, date, service, account, is_from_me)
        """
        fetched_data = self._read_database()

        users = []
        messages = []
        dates = []
        dates_read = []
        service = []
        account = []
        is_from_me = []

        for data in fetched_data:
            users.append(data.user_id)
            messages.append(data.text)
            dates.append(data.date)
            dates_read.append(data.date_read)
            service.append(data.service)
            account.append(data.account)
            is_from_me.append(data.is_from_me)

        data = list(zip(users, messages, dates, dates_read, service, account, is_from_me))

        return data
